/* ************************************************************************
 * Supplemental Material to Brumm and Scheidegger (2017)
 * "Using Adaptive Sparse Grids to Solve High-Dimensional Dynamic Models" 
 * ************************************************************************
 * 
 * This script computes Euler errors for the nonsmooth IRBC model.
 * 
 * It proceeds in three steps:
 * 
 * 1) It reads the policy data that was specified in 
 * filename_policy = "PATH_TO_FILE/FILENAME.txt";
 * 
 * 2) It computes Euler error for random points drawn from the state space.
 * 
 * 3) It reports the errors in 
 *    /results/ERROR.txt (contains summarized results) 
 *    /results/max_error_step.txt (contains the max. error per iteration step)
 * 
 *    and prints the sparse grid coordinates to /results/sparse_grid_plot.plt 
 * 
 *   Notes:
 *   i) the dimension (dim), the number of simulations (no_rand_points)
 *   as well as the ranges for capital (k_up, k_bar) have to be adjusted 
 *   to the model under investigation prior of running the executable.
 *   The values for k_up and k_bar can be found in the respective data 
 *   folders in a file denoted "range.txt".  
 * 
 *   ii) in order to replicate the max. errors reported in the 
 *   paper, one has to compute the 99.th quantile of the max errors 
 *   collected in /results/max_error_step.txt.
 * 
 ************************************************************************** */

#include <mpi.h>
#include <stdlib.h>
#include <random>
#include "incl/SparseGrid.h"
#include "incl/offline_interpol.h"   
#include "incl/parameters.h"

using namespace std;

//!=====================================================================
// mappings from sparse grid to economic domain
//!=====================================================================
//!.....[0,1]^d -> economic domain
extern "C" { double cube_to_box_(double&, double&,int&, double&, 
				 double&, double&, double&, double&);}

//!.....economic domain -> [0,1]^d 				 
extern "C" { double box_to_cube_(double&, double&,int&, double&, 
				 double&, double&, double&, double&);}

//!=====================================================================

//!.....routine to generate Euler Errors 
int simulate();

//!=====================================================================

int main(int argc, char **args) {
  
	//!.....Initialize MPI
	MPI_Init (&argc, &args);
	
	//!.....Call main postprocessing routine to generate Euler Errors 
	simulate();
	
	//!.....Finalize MPI environment.
	MPI_Finalize();

	return 0;
}

//!=====================================================================

int simulate()
{
	//!.....filename of the policies we want to simulate on
	char* filename_policy = "../4d_fix/level3/4d_l3_nonsmooth.txt";

	//!.....create object to simulate Euler errors on
	offline_interpol postprocessing_file;
	
	//!.....load data from file
	postprocessing_file.LoadData(filename_policy);	
	cout << "===================== " << endl;	
	cout << "opening policy data file  " << filename_policy << endl;
	
	//......create filename to write sparse grid to a file
	char* filename_sparsegrid = "results/sparse_grid_plot.plt";	
	cout << "===================== "<< endl;

	//!.....file where errors are written to		
	ofstream reporting_file;
	reporting_file.open("results/ERROR.txt");
	
	//!....this file contains the max. errors per iteration.
	ofstream error_file;
	error_file.open("results/max_error_step.txt");
		
	//!.....grid point -- vector of length dim
	AdaptiveARRAY<double> x_d1;x_d1.redim(dim); 
	 //!.....grid point -- vector of length dim, in [0,1]^d
	AdaptiveARRAY<double> grid_point;grid_point.redim(dim);
	 //!.....grid point -- vector of length dim, in the "economic" domain
	AdaptiveARRAY<double> grid_point_box;grid_point_box.redim(dim);
	//!....next period's state, vector of length dim, in [0,1]^d
	AdaptiveARRAY<double> state_next_cube; state_next_cube.redim(dim);
	//!....next period's state, vector of length dim, in the "economic" domain	
	AdaptiveARRAY<double> state_next; state_next.redim(dim);
	//!....Euler errors for the nonlinear equations	
	AdaptiveARRAY<double> G_eq; G_eq.redim(dim_half+1);  
	//!....auxiliary vars, nonlinear equations	
	AdaptiveARRAY<double> G_eq_aux; G_eq_aux.redim(dim_half+1);  //aux, nonlinear equations
	//!.....init error measures -- average and max errors
	AdaptiveARRAY<double> Max_euler_error;Max_euler_error.redim(dim_half+1);
	AdaptiveARRAY<double> Average_euler_error;Average_euler_error.redim(dim_half+1);
	//!.....init productivity
	AdaptiveARRAY<double> a_plus;a_plus.redim(dim_half);
	
	//!.....initialize errors
	for (int DoF = 1; DoF <= (dim_half+1);DoF++)
	{
	  Max_euler_error(DoF) = 0. ;
	  Average_euler_error(DoF) = 0.;
	}	
	
	//!.....set risk aversion, see Juillard et al. 2011 - JDC, Table 2	
	AdaptiveARRAY<double> gamma; gamma.redim(dim/2);
	for ( int i=1; i<=dim/2; i++)
	{
	  if(dim/2 == 1)
	  {
	    gamma(i) = 0.25;
	  }
	  else
	  {
	    gamma(i) = range_down + (i-1)*range/(dim/2 - 1);
	  }
	}

	//.....Initialize \tau: Negishi weights
	AdaptiveARRAY<double> tau; tau.redim(dim/2);
	double temp_neg;
	for ( int i=1; i<=dim/2; i++)
	{
	  temp_neg = 1/gamma(i);
	  tau(i) = pow(big_A,temp_neg);
	}	
	
	//!.....number of random numbers 
	int no_mc_vars = N;  	
	double r_rand[no_mc_vars];		

	//!.....uniform distributed [0..1]	
	unsigned seed_unif1 = 3;
	std::default_random_engine generator_unif(seed_unif1);
	std::uniform_real_distribution<double> distribution_unif(0.0,1.0);

	//!=====START WITH ERROR COMPUTATION ==================================
	//!.....initialize where to start simulation along a path
	for ( int i=1; i<=dim; i++)
	{
	  x_d1(i) = 0.5;
	}
	//====================================================================
	//!.....begin simulation; from 1 to no_rand_points
	//====================================================================
	for (int random_points = 1; random_points <= no_rand_points; random_points++)
	{ 	
	    //!.....update coordinates
	    for (int i =1; i<=dim; i++)
	    {
	      grid_point(i) = x_d1(i);
	    }
	    
	    //!.....mapping cube [0,1]^d to box [a,b]^d
	    cube_to_box_(grid_point(1), grid_point_box(1), dim,k_bar, a_bar, k_up, a_up, range_cube );
	    
	    //!.....interpolate 
	    postprocessing_file.Interpolate(grid_point,fvalue);
	    double kap[dim_half];
	    double lambda; 
	    for ( int i=0; i< dim_half; i++)
	    {
	      kap[i]=fvalue[i];	      	      
	    }
	    lambda = fvalue[dim_half]; 
	
	    //!.....update the next periods state
	    for (int j = 1; j <= dim_half; j++)
	    {
	      state_next((dim_half) + j) = kap[j-1];
	    }
	    
	    //!.....Monomial integration
	    for (int j = 1; j <= (dim_half+1); j++)
	    {
	      G_eq_aux(j) = 0;  
	    }
 
	    double N_eq =  (dim_half+1);
	    for (int no_mc = 1; no_mc <= 2*N_eq; no_mc++)
	    {  
	      for (int j = 1; j<=N_eq-1; j++)
	      {				
		int_coord = 0.0;
		if (no_mc < 2*N_eq -1)
		{
		  if ((no_mc % 2 == 1) and (no_mc == 2*j -1))  
		  {
		    int_coord = sqrt(N_eq/2.);
		  }
		  else if ((no_mc % 2 == 0) and (no_mc == 2*j))
		  {
		    int_coord = -sqrt(N_eq/2.);
		  }
		}
		else if (no_mc == 2*N_eq -1)  
		{
		  int_coord = sqrt(N_eq/2.);
		}
		else if (no_mc == 2*N_eq)    
		{
		  int_coord = -sqrt(N_eq/2.);
		}
		a_plus(j) = pow(grid_point_box(j),rho)*exp(sigma*(int_coord));
		
		//!.....ensure that productivity is within bounds
		if (a_plus(j) <= a_bar)
		{
		  a_plus(j) = a_bar;
		}
		else if (a_plus(j) >= a_up)
		{
		  a_plus(j) = a_up;
		}
	      }
	 
	      //!.....aux coordinate to interpolate
	      for (int j = 1; j <= dim_half; j++)
	      {
		state_next(j) =  a_plus(j);
	      }
	      
	      //!.....transform state_next -> full coordinate set -> state_next
	      box_to_cube_(state_next(1), state_next_cube(1), dim,k_bar, a_bar, k_up, a_up, range_cube );	            
	      double kap_pp[dim_half]; 
	      double mu_pp[dim_half];
	      postprocessing_file.Interpolate(state_next_cube,fvalue_pp);
		      
	      for ( int i=0; i< dim_half; i++)
	      {
		kap_pp[i]=fvalue_pp[i];
		mu_pp[i] = fvalue_pp[i + dim_half + 1];
	      }
	      lambda_pp = fvalue_pp[dim_half]; 
	      
	      //!.....actual integration <-> summation for N-1 eqs.
	      for (int j = 1; j <= dim_half; j++)
	      {
		double kap_zero = grid_point_box(dim_half + j);
		double mu_plus  = max(0.0, mu_pp[j-1]);
				  
		growth = kap[j-1]/kap_zero -1;
		growth_plus = kap_pp[j-1]/kap[j-1] -1;
		  
		dividend_eq = beta*(lambda_pp*(a_plus(j)*big_A*alpha*pow(kap[j-1],alpha-1) +
				(1-delta) + 0.5*phi* growth_plus*(growth_plus + 2)) - (1.0 -delta)*mu_plus);
		divisor_eq = (lambda*(1+phi*growth));
		quot_eq = dividend_eq / divisor_eq;
		  
		G_eq_aux(j) = G_eq_aux(j) + quot_eq; 				  
		  
	      }    
	    }	    
	    
	  for (int j = 1; j <= dim_half; j++)
	      {    
		double kap_zero = grid_point_box(dim_half + j);
		EE =  - 1 + (1./(2*N_eq))*G_eq_aux(j);
		IC =  1.0 - (kap[j-1])/(kap_zero*(1.-delta));
		G_eq(j) = max(EE,max(IC,min(-EE,-IC)));
		  
		//!.....violations
		if ((kap[j-1]-kap_zero*(1.-delta)) <  epsilon_acc ) 
		{
		  counter_error = counter_error + 1;
		}
		Average_euler_error(j) = Average_euler_error(j) + abs(G_eq(j));
		if (abs(G_eq(j)) >=  Max_euler_error(j))   Max_euler_error(j) =  abs(G_eq(j));
	     }	     

	    for (int j = 1; j <= dim_half; j++)
	      {
		double kap_zero = grid_point_box(dim_half + j);
		growth = kap[j-1]/kap_zero -1;
		
		temp = grid_point_box(j)*big_A*pow(grid_point_box((dim_half)+j),alpha) +
			   grid_point_box((dim_half)+j)*((1 - delta)  - 0.5*phi*(pow(growth,2))) - 
			   kap[j-1]-pow((lambda/tau(j)),-gamma(j)) + temp;
			   
		divisor = grid_point_box(j)*big_A*pow(grid_point_box((dim_half)+j),alpha) +
			  grid_point_box((dim_half)+j)*(-0.5*phi*(pow(growth,2))) + divisor;			   
	     }
	     
	    temp_sum = temp/divisor;
	    G_eq(dim_half+1) = temp_sum;     
	    Average_euler_error(dim_half+1) = Average_euler_error(dim_half+1) + (abs(G_eq(dim_half+1)));
	    if ((abs(G_eq(dim_half+1))) >=  Max_euler_error(dim_half+1))   Max_euler_error(dim_half+1) =  (abs(G_eq(dim_half+1)));

	    //!......average of max per point
	    double max_max_dummy = -10;
	    for (int j = 1; j <= dim_half + 1; j++)
	    {
	      if (abs(G_eq(j)) >=  max_max_dummy)   max_max_dummy = abs(G_eq(j));
	    }
	    ave_max = ave_max + max_max_dummy;
	    
	    if (verbosity == 1)
	      cout << "===point no. " <<random_points<<" ===========" << endl;	

	    //!...... draw random numbers in [0..1] 	    
	    for ( int i=1; i<=dim; i++)
	    {
	      double MC_double  = distribution_unif(generator_unif);
	      x_d1(i) = MC_double;
	      x_d1(i) = MC_double*range_MC + range_down_MC; 
	    }    
	
	  double max_time_error = 0;
	  for (int j = 1; j <= dim_half + 1; j++)
	  {
 	    if (abs(G_eq(j)) >=  max_time_error)   max_time_error = abs(G_eq(j));
	  }
	  error_file << "max error in timestep " << random_points << "  " << log10(max_time_error) <<  endl;
	} 
	//====================================================================	     
	//!========= end loop over sample points==============================
	//====================================================================	     
	for (int j = 1; j <= dim_half+1; j++)
	{
	  cout<< "Max EULER PATH error per DoF " << j << "  Max error " << log10(Max_euler_error(j)) << endl;
	  cout<< "Average EULER PATH error per DoF " << j << "  Average error   "<< log10(Average_euler_error(j)/(no_rand_points)) << endl;
	  reporting_file << "Max EULER PATH error per DoF " << j << "  Max error " << log10(Max_euler_error(j)) << endl;
	  reporting_file << "Average EULER PATH error per DoF " << j << "  Average error   "<< log10(Average_euler_error(j)/(no_rand_points)) << endl;
	}	
	
	for (int j = 1; j <= dim_half + 1; j++)
	{
 	    if (Max_euler_error(j) >=  max_max_error)   max_max_error = Max_euler_error(j);
	    ave_ave = log10(Average_euler_error(j) / no_rand_points) + ave_ave;
	}
	cout << "===================== "<< endl;
	cout << "REPORTING: MAX ERROR " << log10(max_max_error) << endl;
	cout << "REPORTING: AVERAGE MAX ERROR " << log10(ave_max/(no_rand_points)) << endl;
	cout << "REPORTING: AVERAGE AVERAGE ERROR " << ave_ave/(dim_half + 1) << endl;
		
	reporting_file << "===================== "<< endl;
	reporting_file << "REPORTING: MAX ERROR " << log10(max_max_error) << endl;
	reporting_file << "REPORTING: AVERAGE MAX ERROR " << log10(ave_max/(no_rand_points)) << endl;
	reporting_file << "REPORTING: AVERAGE AVERAGE ERROR " << ave_ave/(dim_half + 1) << endl;
	
	//!.....print sparse grid
	postprocessing_file.PlotSparseGrid(filename_sparsegrid);
	
	//!.....close files
	reporting_file.close();
	error_file.close();
			
//!=====================================================================
      return 1;
}  
//!=====================================================================	  
  

  
  


