/* ************************************************************************
 * Supplemental Material to Brumm and Scheidegger (2017)
 * "Using Adaptive Sparse Grids to Solve High-Dimensional Dynamic Models" 
 * ************************************************************************
 * 
 * This script computes Euler errors for the smooth IRBC model.
 * 
 * It proceeds in three steps:
 * 
 * 1) main calls the routine simulate(). The latter reads the 
 *    policy data that was specified in 
 *    filename_policy = "PATH_TO_FILE/FILENAME.txt".
 * 
 * 2) It computes Euler errors along a simulated path.
 * 
 * 3) It reports the errors in 
 *    /results/ERROR.txt (contains summarized results)  
 *    /results/simulated_path.txt (simulated paths)
 * 
 *    and prints the sparse grid coordinates to /results/sparse_grid_plot.plt 
 * 
 *   Notes:
 *   i) the dimension (dim), the number of simulations (no_rand_points)
 *   as well as the ranges for capital (k_up, k_bar) have to be adjusted 
 *   to the model under investigation prior of running the executable.
 *   The values for k_up and k_bar can be found in the respective data 
 *   folders in a file denoted "range.txt".
 * 
 ************************************************************************** */

#include <mpi.h>
#include <stdlib.h>
#include <random>
#include "incl/SparseGrid.h"
#include "incl/offline_interpol.h"   
#include "incl/parameters.h"

using namespace std;

//!=====================================================================
// mappings from sparse grid to economic domain
//!=====================================================================
//!.....[0,1]^d -> economic domain
extern "C" {
	double cube_to_box_(double&, double&, int&, double&,
	                    double&, double&, double&, double&);
}

//!.....economic domain -> [0,1]^d 
extern "C" {
	double box_to_cube_(double&, double&, int&, double&,
	                    double&, double&, double&, double&);
}

//!=====================================================================

//!.....routine to generate Euler Errors
int simulate();

//!=====================================================================

int main(int argc, char **args) {
  
	//!.....Initialize MPI
	MPI_Init (&argc, &args);
	
	//!.....Call main postprocessing routine to generate Euler Errors 
	simulate();
	
	//!.....Finalize MPI environment.
	MPI_Finalize();

	return 0;
}

//----------------------------------------------------------------------

int simulate() {

	//!.....filename of the policies we want to simulate on
	char* filename_policy = "../4d/level3/4d_l3_smooth.txt";
	
	//!.....create object to simulate Euler errors on
	offline_interpol postprocessing_file;
	
	//!.....load data from file
	postprocessing_file.LoadData(filename_policy);	
	cout << "===================== " << endl;	
	cout << "opening policy data file  " << filename_policy << endl;

	//!.....create filename to write sparse grid to a file
	char* filename_sparsegrid = "results/sparse_grid_plot.plt";

	//!.....file where errors are written to	
	ofstream reporting_file;
	reporting_file.open("results/ERROR.txt");	
	
	ofstream path_file;
	path_file.open("results/simulated_path.txt");	
	
	//!.....grid point -- vector of length dim
	AdaptiveARRAY<double> x_d1; x_d1.redim(dim);
	 //!.....grid point -- vector of length dim, in [0,1]^d
	AdaptiveARRAY<double> grid_point; grid_point.redim(dim);
	 //!.....grid point -- vector of length dim, in the "economic" domain
	AdaptiveARRAY<double> grid_point_box; grid_point_box.redim(dim);	
	//!....next period's state, vector of length dim, in [0,1]^d
	AdaptiveARRAY<double> state_next_cube; state_next_cube.redim(dim);
	//!....next period's state, vector of length dim, in the "economic" domain
	AdaptiveARRAY<double> state_next; state_next.redim(dim);
	//!....Euler errors for the nonlinear equations
	AdaptiveARRAY<double> G_eq; G_eq.redim(N);
	//!....auxiliary vars, nonlinear equations	
	AdaptiveARRAY<double> G_eq_aux; G_eq_aux.redim(N);  
	//!.....init error measures -- average and max errors
	AdaptiveARRAY<double> Max_euler_error; Max_euler_error.redim(N);
	AdaptiveARRAY<double> Average_euler_error; Average_euler_error.redim(N);
	//!.....init productivity
	AdaptiveARRAY<double> a_plus; a_plus.redim(N - 1);	
	//!.....init productivity to simulate
	AdaptiveARRAY<double> a_plus_point; a_plus_point.redim(N - 1);
	
	//!.....init errors...
	for (int DoF = 1; DoF <= N; DoF++) {
		Max_euler_error(DoF) = -10. ;
		Average_euler_error(DoF) = 0.;
	}	
	
	//!.....initialize risk aversion, see Juillard et al. 2011 - JDC, Table 2
	AdaptiveARRAY<double> gamma; gamma.redim(dim / 2);
	for ( int i = 1; i <= dim / 2; i++) {
		if (dim / 2 == 1) {
			gamma(i) = 5. / 8.;
		} else {
			gamma(i) = range_down + (i - 1) * range / (dim / 2 - 1);
		}
	}
	//!.....Initialize \tau: Negishi weights
	AdaptiveARRAY<double> tau; tau.redim(dim / 2);

	double temp_neg;
	for ( int i = 1; i <= dim / 2; i++) {
		temp_neg = 1 / gamma(i);
		tau(i) = pow(big_A, temp_neg);
	}	
	
	//!.....number of random numbers
	int no_mc_vars = N;  	
	double r_rand[no_mc_vars];

	//!.....random number generator 
	unsigned seed_m = 299;
	std::default_random_engine generator_m(seed_m);
	std::normal_distribution<double> distribution_m(0.0, 1.0);

	//!=====START WITH ERROR COMPUTATION ==================================
	//!.....initialize where to start simulation along a path
	for ( int i = 1; i <= dim; i++) {
		x_d1(i) = 0.5;//centre of [0,1]^d cube centre
	}

	//====================================================================
	//!.....begin simulation; from 1 to no_rand_points
	//====================================================================
	for (int random_points = 1; random_points <= no_rand_points; random_points++) {
		
		//!.....update coordinates
		for (int i = 1; i <= dim; i++) {
			grid_point(i) = x_d1(i);
		}

		//!.....mapping cube [0,1]^d to box [a,b]^d
		cube_to_box_(grid_point(1), grid_point_box(1), dim, k_bar, a_bar, k_up, a_up, range_cube );

		//!.....write for every simulation step the state to file
		path_file  << "  " << random_points ;
		for ( int i = 1; i <= dim; i++) {
			path_file << "  " << grid_point_box(i) << " " ;
		}
		path_file << "  " << endl;

		//!.....interpolate 
		postprocessing_file.Interpolate(grid_point, fvalue);
	
		double kap[N - 1]; 
		double lambda;   
		for ( int i = 0; i < N - 1; i++) {
			kap[i] = fvalue[i];
		}
		lambda = fvalue[N - 1];

		//!.....update the next periods state
		for (int j = 1; j <= N - 1; j++) {
			state_next((N - 1) + j) = kap[j - 1];
		}

		//!.....initialize integration value
		for (int j = 1; j <= N; j++) {
			G_eq_aux(j) = 0;  
		}

		//!.....monomial integration
		for (int no_mc = 1; no_mc <= 2 * N; no_mc++) {
			for (int j = 1; j <= N - 1; j++) {
				int_coord = 0.0;
				if (no_mc < 2 * N - 1) {
					if ((no_mc % 2 == 1) and (no_mc == 2 * j - 1)) { 
						int_coord = sqrt(N / 2.);
					} else if ((no_mc % 2 == 0) and (no_mc == 2 * j)) {
						int_coord = -sqrt(N / 2.);
					}
				} else if (no_mc == 2 * N - 1) { 
					int_coord = sqrt(N / 2.);
				} else if (no_mc == 2 * N) { 
					int_coord = -sqrt(N / 2.);
				}
				a_plus(j) = pow(grid_point_box(j), rho) * exp(sigma * (int_coord));
				//!.....ensure that productivity is within bounds
				if (a_plus(j) <= a_bar) {
					a_plus(j) = a_bar;
				} else if (a_plus(j) >= a_up) {
					a_plus(j) = a_up;
				}
			}

			//!.....aux coordinate to interpolate
			for (int j = 1; j <= N - 1; j++) {
				state_next(j) =  a_plus(j);
			}

			//!.....transform state_next -> full coordinate set -> state_next
			box_to_cube_(state_next(1), state_next_cube(1), dim, k_bar, a_bar, k_up, a_up, range_cube );

			double lambda_pp;
			double kap_pp[N - 1]; 
			postprocessing_file.Interpolate(state_next_cube, fvalue_pp); 

			for ( int i = 0; i < N - 1; i++) {
				kap_pp[i] = fvalue_pp[i];
			}
			lambda_pp = fvalue_pp[N - 1];

			//!.....actual integration: summation for N-1 eqs.
			for (int j = 1; j <= N - 1; j++) {
				double kap_zero = grid_point_box(N - 1 + j);

				growth = kap[j - 1] / kap_zero - 1;
				growth_plus = kap_pp[j - 1] / kap[j - 1] - 1;

				dividend_eq = beta * (lambda_pp * (a_plus(j) * big_A * alpha * pow(kap[j - 1], alpha - 1) +
				                                   (1 - delta) + 0.5 * phi * growth_plus * (growth_plus + 2)));
				divisor_eq = (lambda * (1 + phi * growth));
				quot_eq = dividend_eq / divisor_eq;

				G_eq_aux(j) = G_eq_aux(j) + quot_eq;

			}  
		  }
		  
		//!.....first N-1 nonlinear equations, MC integration -> average
		for (int j = 1; j <= N - 1; j++) {
			G_eq(j) = 1 - (1. / (2 * N)) * G_eq_aux(j);
			//euler errors
			Average_euler_error(j) = Average_euler_error(j) + abs(G_eq(j));
			if (abs(G_eq(j)) >=  Max_euler_error(j))   { Max_euler_error(j) =  abs(G_eq(j)); }
		}
	
		//!.....last nonlinear equation
		double temp = 0;
		double temp_sum = 0;
		double divisor = 0;
		for (int j = 1; j <= N - 1; j++) {
			double kap_zero = grid_point_box(N - 1 + j);
			growth = kap[j - 1] / kap_zero - 1;

			temp = grid_point_box(j) * big_A * pow(grid_point_box((N - 1) + j), alpha) +
			       grid_point_box((N - 1) + j) * ((1 - delta)  - 0.5 * phi * (pow(growth, 2))) -
			       kap[j - 1] - pow((lambda / tau(j)), -gamma(j)) + temp;

			divisor = grid_point_box(j) * big_A * pow(grid_point_box((N - 1) + j), alpha) +
			          grid_point_box((N - 1) + j) * (-0.5 * phi * (pow(growth, 2))) + divisor;
		}

		temp_sum = temp / divisor;
		G_eq(N) = temp_sum; 

		Average_euler_error(N) = Average_euler_error(N) + abs(G_eq(N));
		if (abs(G_eq(N)) >=  Max_euler_error(N))   { Max_euler_error(N) =  abs(G_eq(N)); }

		//!.....average of max per point
		double max_max_dummy = -10;
		for (int j = 1; j <= N; j++) {
			if (abs(G_eq(j)) >=  max_max_dummy)   { max_max_dummy = abs(G_eq(j)); }
		}
		ave_max = ave_max + max_max_dummy;

		if (verbosity == 1)
		    cout << " simulation point no. " << random_points <<  endl;

		//!.....generate random numbers ~N(0,1) for productivity
		for (int j = 0; j < N; j++) {
			r_rand[j] = distribution_m(generator_m);
		}

		double rand_point_fix = r_rand[0];
		for (int j = 1; j <= N - 1; j++) {
			double rand_point_fix_country;
			rand_point_fix_country = r_rand[j];
			a_plus_point(j) = pow(grid_point_box(j), rho) * exp(sigma * (rand_point_fix + rand_point_fix_country));
			//!.....ensure that productivity is within bounds
			if (a_plus_point(j) <= a_bar) {
				a_plus_point(j) = a_bar;
			} else if (a_plus_point(j) >= a_up) {
				a_plus_point(j) = a_up;
			}
		}

		//!.....update grid points
		for ( int i = 1; i <= dim / 2; i++) {
			grid_point_box(i) = a_plus_point(i);
			grid_point_box(dim / 2 + i) = kap[i - 1];
		}

		//!.....transform to [0,1]^d
		box_to_cube_(grid_point_box(1), x_d1(1), dim, k_bar, a_bar, k_up, a_up, range_cube );

	} 
	//====================================================================	     
	//!========= end loop over sample points==============================
	//====================================================================		
	
	
	cout << "===================== " << endl;
	for (int j = 1; j <= N; j++) {
		cout << "Max EULER PATH error per DoF " << j << "  Max error " << log10(Max_euler_error(j)) << endl;
		cout << "Average EULER PATH error per DoF " << j << "  Average error   " << log10(Average_euler_error(j) / no_rand_points) << endl;
		reporting_file << "Max EULER PATH error per DoF " << j << "  Max error " << log10(Max_euler_error(j)) << endl;
		reporting_file << "Average EULER PATH error per DoF " << j << "  Average error   " << log10(Average_euler_error(j) / no_rand_points) << endl;
	}

	for (int j = 1; j <= N; j++) {
		if (Max_euler_error(j) >=  max_max_error)   { max_max_error = Max_euler_error(j); }//max error
		ave_ave = log10(Average_euler_error(j) / no_rand_points) + ave_ave;
	}
	cout << "===================== " << endl;
	cout << "REPORTING: MAX ERROR " << log10(max_max_error) << endl;
	cout << "REPORTING: AVERAGE MAX ERROR " << log10(ave_max / no_rand_points) << endl;
	cout << "REPORTING: AVERAGE AVERAGE ERROR " << ave_ave/N << endl;

	reporting_file << "===================== " << endl;
	reporting_file << "REPORTING: MAX ERROR " << log10(max_max_error) << endl;
	reporting_file << "REPORTING: AVERAGE MAX ERROR " << log10(ave_max / no_rand_points) << endl;
	reporting_file << "REPORTING: AVERAGE AVERAGE ERROR " << ave_ave/N << endl;

	//!.....print sparse grid
	postprocessing_file.PlotSparseGrid(filename_sparsegrid);

	//!.....close file	
	reporting_file.close();
	path_file.close();
	
//!=====================================================================
	return 1;
}  
//!=====================================================================






