%**************************************************************************
% Supplemental Material to Brumm and Scheidegger (2017)
% "Using Adaptive Sparse Grids to Solve High-Dimensional Dynamic Models" 
%**************************************************************************
% LAW OF MOTION 
%**************************************************************************
% This script calculates the law of motion and consumption moments for 
% the menu costs models in Brumm and Scheidegger (2017). 
% It builds on supplemental material to Midrigan (2011). 
% It proceeds in three steps: 
% 1) Read data (price level, money growth).
% 2) Find coefficients for law of motion.
% 3) Calculate consumption moments.
%**************************************************************************

clear all; clc; 

%% 1) Read data (price level, money growth):
max_rank=7; 
drop=25;
logoldprice=[];lognewprice=[];logmoneygr=[];
for count_rank=0:max_rank % consider (max_rank+1) simultions
T = readtable(['2goods_data_orig\Aggregates.rank_',num2str(count_rank),'.txt'],'Delimiter',' ','ReadVariableNames',false);
%T = readtable(['3goods_data_orig\Aggregates.rank_',num2str(count_rank),'.txt'],'Delimiter',' ','ReadVariableNames',false);
logprice = log(T{:,{'Var2'}});
logprice = logprice(drop:375);
logmoneygrowth = log(T{:,{'Var3'}});
logmoneygrowth = logmoneygrowth(drop:375);
logoldprice = [logoldprice; logprice(1:end-1)];
lognewprice = [lognewprice; logprice(2:end)];
logmoneygr = [logmoneygr; logmoneygrowth(1:end-1)];
end

%% 2) Find coefficients for law of motion:
y = lognewprice ;
x = [logmoneygr logoldprice];
n = length(x);
X = [ones(n, 1) x]; 
beta_hat = (X'*X)\X'*y;
zeta=beta_hat';
display(['Zeta:     ',num2str(zeta)]) 


%% 3) Calculate consumption moments:
T = 1000000;
C = zeros(T,1);
g = zeros(T,1);
a1 = zeta(3);
a2 = zeta(2);
rog  = 0.61^(1/4);      % consistent with earlier VAR evidence
seg  = sqrt(0.0018^2/((1+rog^6)+(1+rog^4)*(1+rog)^2+(1+rog^2)*(1+rog+rog^2)^2+(1+rog+rog^2+rog^3)^2));

for t = 2:T
    g(t) = rog*g(t-1)+randn*seg;
    C(t) = a1*C(t-1)+a2*g(t); 
end

Cm = reshape(C',4,T/4);   Cm = exp(Cm); Cm = sum(Cm)';   Cm = log(Cm/4);
[Cd,Ct] = hpfilter(Cm,14400);

display(['STD of Consumption in Percent: ',num2str(std(Ct)*100)]) 
display(['Serial Correl. of Consumption: ',num2str(corr(Ct(1:end-1),Ct(2:end)))]) 