/* ************************************************************************
 * Supplemental Material to Brumm and Scheidegger (2017)
 * "Using Adaptive Sparse Grids to Solve High-Dimensional Dynamic Models" 
 * ************************************************************************
 * 
 * This script simulates the menu cost model of Section 5 for many firms 
 * to obtain aggregates.
 * 
 * It proceeds in three steps:
 * 
 * 1) main calls the routine simulate(). The latter reads the 
 *    sparse grid value function data that was specified in the two files 
 *    "2goods_data/MC_2goods_state_1.txt";
 *    "2goods_data/MC_2goods_state_2.txt";
 * 
 * 2) It simulates the MC model.
 * 
 * 3) It reports the time series for the aggregate price level and money 
 *    growth in:
 *    "results/Aggregates.MPI_RANK.txt"; 
 * 
 ************************************************************************** */

#include <mpi.h>
#include <stdlib.h>
#include <random>
#include "incl/SparseGrid.h"
#include "incl/offline_interpol.h"   
#include "incl/parameters.h"

using namespace std;

//!=====================================================================
// mappings from sparse grid to economic domain
//!=====================================================================
//!.....[0,1]^d -> economic domain

extern "C" { 
  double box_to_cube_mapping_(double&, double&, int&, double&, double&, double&,
	                    double&, double&, double&);

}


//!=====================================================================

//!.....routine to simulate aggregates
int simulate();

//!=====================================================================


int main(int argc,char **args)
{
	//Initialize MPI
	MPI_Init (&argc, &args);
	{
  
	  
//!.....MPI test........................................................  
	  int numprocs, rank, namelen;
	  char processor_name[MPI_MAX_PROCESSOR_NAME];
	  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);			
	  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	  MPI_Get_processor_name(processor_name, &namelen);

	
//!.......check how many processors are there .........................	
	  cout << "hello from Process rank # " << rank <<" in the communicator with "  << numprocs  << " processes in total " << endl;  
	  MPI_Barrier(MPI_COMM_WORLD);
	  
//!.......Call main postprocessing routine.............................
	  simulate();
	}

	MPI_Finalize(); //Finalize MPI 

	return 0;
}


//!=====================================================================
//!=====================================================================

int simulate()
{
      
//!------------------------------------------------------------- 

	//!.....filename of the data we want to simulate on
	char* filename1 = "2goods_data/MC_2goods_state_1.txt";
	char* filename2 = "2goods_data/MC_2goods_state_2.txt";

	//!.....MPI 
	int rank;
	int numprocs;
 	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);	
	
	//!.....Load data objects
	offline_interpol interpolate_fun_state_1, interpolate_fun_state_2; 
	interpolate_fun_state_1.LoadData(filename1); 
	cout << "opening data file  "<< filename1 << " on rank " << rank  <<  endl; 
	interpolate_fun_state_2.LoadData(filename2); 
	cout << "opening data file  "<< filename2 <<  " on rank " << rank  <<  endl; 
	
	//......path to results
	char* nameRes3 = "results/Aggregates"; 
	char filename_pco[255];
	sprintf(filename_pco,"%s.rank_%d.txt", nameRes3,rank);	
	
	//!.....open output
	ofstream pco_file;
	pco_file.open(filename_pco);	
	
	//!....next period's state, vector of length dim, in [0,1]^d
	AdaptiveARRAY<double> state_next_box; state_next_box.redim(dim);
	AdaptiveARRAY<double> state_next_cube; state_next_cube.redim(dim);
	//!.....state_next_box per firm, entires 1 and 2
	AdaptiveARRAY<double> state_next_box_1; state_next_box_1.redim(nrfirms);
	AdaptiveARRAY<double> state_next_box_2; state_next_box_2.redim(nrfirms);	
	AdaptiveARRAY<double> mu_co; mu_co.redim(dim/2);
	//!.....one mu per firm	
	AdaptiveARRAY<double> mu_temp; mu_temp.redim(dim/2);
	AdaptiveARRAY<double> mu_temp_1; mu_temp_1.redim(nrfirms);
	AdaptiveARRAY<double> mu_temp_2; mu_temp_2.redim(nrfirms);
	//!.....one mu per firm	
	AdaptiveARRAY<double> mu_temp_old; mu_temp_old.redim(dim/2);
	AdaptiveARRAY<double> mu_temp_old_1; mu_temp_old_1.redim(nrfirms);
	AdaptiveARRAY<double> mu_temp_old_2; mu_temp_old_2.redim(nrfirms);	
	//!.....one mu per firm		
	AdaptiveARRAY<double> mu_reg; mu_reg.redim(dim/2);
	AdaptiveARRAY<double> mu_reg_1; mu_reg_1.redim(nrfirms);
	AdaptiveARRAY<double> mu_reg_2; mu_reg_2.redim(nrfirms);	
	//!.....one mu per firm			
	AdaptiveARRAY<double> mu_reg_old; mu_reg_old.redim(dim/2);
	AdaptiveARRAY<double> mu_reg_old_1; mu_reg_old_1.redim(nrfirms);
	AdaptiveARRAY<double> mu_reg_old_2; mu_reg_old_2.redim(nrfirms);		
	//!.....one mu per firm		
	AdaptiveARRAY<double> mu_co_next; mu_co_next.redim(dim/2);
	AdaptiveARRAY<double> mu_co_next_1; mu_co_next_1.redim(nrfirms);
	AdaptiveARRAY<double> mu_co_next_2; mu_co_next_2.redim(nrfirms);		
	AdaptiveARRAY<double> e_current;e_current.redim(nrfirms);
	AdaptiveARRAY<int> state_in;state_in.redim(nrfirms);	

//!.....normally distributed number generators
	unsigned seed_m = 1 + 5*(rank + 1);
	std::default_random_engine generator_m(seed_m); 
	std::normal_distribution<double> distribution_m(0.0,sigma_m);

	unsigned seed_a = 2+ 5*(rank + 1);
	std::default_random_engine generator_a(seed_a); 
	std::normal_distribution<double> distribution_a(0.0,sigma_a);	
	
	unsigned seed_a2 = 22+ 5*(rank + 1);
	std::default_random_engine generator_a2(seed_a2); 
	std::normal_distribution<double> distribution_a2(0.0,sigma_a);		
	
//!.....uniform distributed [0..1]	
	unsigned seed_unif1 = 3 + 5*(rank + 1);
	std::default_random_engine generator_unif(seed_unif1);
	std::uniform_real_distribution<double> distribution_unif(0.0,1.0);

	unsigned seed_unif2 = 4 + 5*(rank + 1);
	std::default_random_engine generator_unif_2(seed_unif2);
	std::uniform_real_distribution<double> distribution_unif_2(0.0,1.0);	
	
	unsigned seed_unif3 = 5 + 5*(rank + 1);
	std::default_random_engine generator_rand_unif(seed_unif3);
	std::uniform_real_distribution<double> distribution_rand_unif(0.0,1.0);			
	
//!.....uniform distributed [0..1]	
	unsigned seed_unif_state = 666 + 5*(rank + 1);
	std::default_random_engine generator_unif_state(seed_unif_state);
	std::uniform_real_distribution<double> distribution_unif_state(0.0,1.0);
	double uniform_state  = distribution_unif_state(generator_unif_state);	
	for (int j = 1; j <=nrfirms; j++){
	  uniform_state  = distribution_unif_state(generator_unif_state);
	  if(uniform_state <= 0.21)
	  {
	      state_in(j) = 1;}  
	   else
	   {
	      state_in(j) = 2; 
	   }
	}  	
		
	//!.....set initial point point
	for(int firm_count = 1; firm_count <= nrfirms; firm_count++)
	{
	  state_next_box_1(firm_count) = 1.5;
	  state_next_box_2(firm_count) = 1.5;
	} 
	state_next_box(1) = 1.5;
	state_next_box(2) = 1.5;
	state_next_box(3) = 0.5*(g_up + g_down);
	state_next_box(4) = 0.5*(p_up + p_down);

	//!.....transform to cube [0..1]^d
	box_to_cube_mapping_(state_next_box(1), state_next_cube(1),dim, mu_up, mu_down, g_up, g_down, p_up, p_down);

	//!.....Initialize price change parameters
    p1_old = 1.0; 
    eta_a1_old  = 1.0;
	eta_a2_old  = 1.0;
	
	e_plus_old  = 1.0;
	
	mu_temp_old(1) = 1.0;
	mu_temp_old(2) = 1.0; 
	
	mu_reg(1) = 1.0;
	mu_reg(2) = 1.0;
	
    mu_reg_old(1) = 1.0;
	mu_reg_old(2) = 1.0;

	for (int j = 1; j <=nrfirms; j++){
	  mu_temp_old_1(j) = 1; 
	  mu_temp_old_2(j) = 1;
	  mu_reg_1(j) = 1;
	  mu_reg_2(j) = 1;
	  mu_reg_old_1(j) = 1;
	  mu_reg_old_2(j) = 1;
	}
	
	
	//!=====================================================================	
	//!.....Major simulation loop
	//!=====================================================================	      
	//!.....state counter
	int state_sum_1 =0, state_sum_2=0;        
        
	for(int loop_count = 1; loop_count <=nrsim; loop_count++)
	{ 
	  for(int nfirm_count = 1; nfirm_count<= nrfirms; nfirm_count++)
	  {
	    
	   if(loop_count ==1)
	   {
	      state_next_box(1) = 1.5;
	      state_next_box(2) = 1.5;
	      state_next_box(3) = 0.5*(g_up + g_down);
	      state_next_box(4) = 0.5*p_up + 0.5*p_down;	     
	      box_to_cube_mapping_(state_next_box(1), state_next_cube(1),dim, mu_up, mu_down, g_up, g_down, p_up, p_down);
	   }
	   
	   else 
	   {
    	      state_next_box(1) = state_next_box_1(nfirm_count);
	      state_next_box(2) = state_next_box_2(nfirm_count);     
	      box_to_cube_mapping_(state_next_box(1), state_next_cube(1),dim, mu_up, mu_down, g_up, g_down, p_up, p_down);
	   }

	  //!.....in which state are we now - interpolant
	  if(state_in(nfirm_count) == 1)
	  {
	    interpolate_fun_state_1.Interpolate(state_next_cube,fvalue);
	    VF1 = fvalue[0];
	    mu_R = fvalue[1];
	    VF2 = fvalue[2];
	    mu = fvalue[3];
	    VF3 = fvalue[4];
	    state_sum_1 = state_sum_1 + 1;
	  }

	  else if (state_in(nfirm_count) == 2)
	  {
	    interpolate_fun_state_2.Interpolate(state_next_cube,fvalue);
	    VF1 = fvalue[0];
	    mu_R = fvalue[1];
	    VF2 = fvalue[2];
	    mu = fvalue[3];
	    VF3 = fvalue[4];
    	    state_sum_2 = state_sum_2 + 1;
	  }
	  
//!.......value function choices	  
	  if (VF1 > max(VF2,VF3))
	  {
	     x1s = mu_R;
	     x2s = mu_R;
	     mu_temp_1(nfirm_count) = mu_R;
	     mu_temp_2(nfirm_count) = mu_R;
	     mu_reg_1(nfirm_count) = mu_R;
	     mu_reg_2(nfirm_count) = mu_R;
	     price_disc = 1.0;
	     price_cont = mu_R;
	     price_1 = price_1 + 1; 	    
	  }
	  else if (VF2 > max(VF1,VF3))
	  {
    	     x1s = state_next_box(1);
	     x2s = state_next_box(2);
	     mu_temp_1(nfirm_count) = mu;
	     mu_temp_2(nfirm_count) = mu;
	     mu_reg_1(nfirm_count) = state_next_box(1);
	     mu_reg_2(nfirm_count) = state_next_box(2);
	     price_disc = 2.0;
	     price_cont = mu;
	     price_2 = price_2 + 1;
	  }
	  else if (VF3 > max(VF1,VF2))
	  {
       	     x1s = state_next_box(1);
	     x2s = state_next_box(2);
	     mu_temp_1(nfirm_count) = state_next_box(1);
	     mu_temp_2(nfirm_count) = state_next_box(2);
	     mu_reg_1(nfirm_count) = state_next_box(1);
	     mu_reg_2(nfirm_count) = state_next_box(2);
	     price_disc = 3.0;
	     price_cont = 0.0;
     	     price_3 = price_3 + 1;
	  }
	  
	  //!.....update coords
	  mu_co(1) = x1s; 
	  mu_co(2) = x2s;
	  g_co = state_next_box(3);
	  p_co = state_next_box(4);
	   
	  //!.....draw random numbers -- draw only once per timestep, same for all firms
	  if(nfirm_count == 1){
  	  eta_m = distribution_m(generator_m);
	  }
	  
	  double eta_a1_normal = distribution_a(generator_a);
	  double eta_a2_normal = distribution_a2(generator_a2);
	  double eta_a1_tilde;
	  double eta_a2_tilde;
	  double eta_a1_uniform  = distribution_unif(generator_unif);
	  double eta_a2_uniform  = distribution_unif_2(generator_unif_2);
	  
	  //!.....first random var
	  if(eta_a1_uniform <= p_a)
	  {
	    eta_a1_tilde = eta_a1_normal;
	  }
	  else 
	  {
	   eta_a1_tilde = 0.0; 
	  }
	  
	   //!.....second random var
	  if(eta_a2_uniform <= p_a)
	  {
	    eta_a2_tilde = eta_a2_normal;
	  }
	  else 
	  {
	   eta_a2_tilde = 0.0; 
	  }
	  
	  double eta_mean = 0.5*(eta_a1_tilde + eta_a2_tilde); 
	  eta_a1 = eta_a1_tilde + xi*eta_mean;
	  eta_a2 = eta_a2_tilde + xi*eta_mean;
	  	  
	  //!.....get e_current
	 if(state_in(nfirm_count) == 1)
	 {
	  e_current(nfirm_count) = e_bar; 
	 }
	 else if (state_in(nfirm_count) == 2)
	 {
	  e_current(nfirm_count) = 1.0;
	 }	  

	  //!.....draw random number in [0..1]
	  double rand_unif = distribution_unif(generator_rand_unif); 
	  
	  //!.....get e_next and update state for next period
	 if(state_in(nfirm_count) == 1)
	 {
	  prob_v1 = rho;
	  prob_v2 = 1.0 -rho;
	  if(rand_unif < prob_v1)
	  {
	    state_plus1 = 1;
	    e_next = e_bar;
	  }
	  else if(rand_unif >= prob_v1)
	  {
	    state_plus1 = 2;
	    e_next = 1;
	  }
	 }
	 else if (state_in(nfirm_count) == 2)
	 {
          prob_v1 = alpha;
	  prob_v2 = 1.0 -alpha;
	  if(rand_unif < prob_v1)
	  {
	    state_plus1 = 1;
	    e_next = e_bar;
	  }
	  else if(rand_unif >= prob_v1)
	  {
	    state_plus1 = 2;
	    e_next = 1;
	  }
	 }
	  
	//!.....compute price changes p(t) / p(t-1)
	delta_a1_plus_old  = eta_a1_old;
	delta_a2_plus_old  = eta_a2_old;
	Gplus  = state_next_box(3);
	e_plus_current  = e_current(nfirm_count);

	//!.....price change		
	delta_price  = (mu_temp_1(nfirm_count)/mu_temp_old_1(nfirm_count))*exp(delta_a1_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	delta_price2 = (mu_temp_2(nfirm_count)/mu_temp_old_2(nfirm_count))*exp(delta_a2_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	
	//!.....regular price change		
	delta_price_R  = (mu_reg_1(nfirm_count)/mu_reg_old_1(nfirm_count))*exp(delta_a1_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	delta_price2_R = (mu_reg_2(nfirm_count)/mu_reg_old_2(nfirm_count))*exp(delta_a2_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	
	//!.....update old values	
	e_plus_old  = e_plus_current;	
	eta_a1_old = eta_a1; 
    eta_a2_old = eta_a2;
	
    mu_temp_old_1(nfirm_count) = mu_temp_1(nfirm_count);
	mu_temp_old_2(nfirm_count) = mu_temp_2(nfirm_count);
	
	mu_reg_old_1(nfirm_count) = mu_reg_1(nfirm_count); 
	mu_reg_old_2(nfirm_count) = mu_reg_2(nfirm_count);
	
	//!.....update state + grid point here...         	  
	state_in(nfirm_count) = state_plus1;

	//!.....only update once per timestep -- same for all firms
	if(nfirm_count == 1){
	  if(with_m_shock == 0)
	  {
	    eta_m = 0.0; //debug flag
	  }
 	g_co_next = exp(rho_m*log(g_co) + eta_m);}
	
 	mu_co_next_1(nfirm_count) = (mu_co(1)*e_current(nfirm_count))/(g_co_next*exp(eta_a1)*e_next);
 	mu_co_next_2(nfirm_count) = (mu_co(2)*e_current(nfirm_count))/(g_co_next*exp(eta_a2)*e_next);
	
	state_next_box(1) = mu_co_next_1(nfirm_count);
	state_next_box(2) = mu_co_next_2(nfirm_count);
	
	//!.....backup -- force simulation into the box
	if(state_next_box(1) <= mu_down)
	      state_next_box(1)= mu_down;
	if(state_next_box(1) >= mu_up)
	      state_next_box(1) = mu_up;
	  
	if(state_next_box(2) <= mu_down)
	      state_next_box(2)= mu_down;
	if(state_next_box(2) >= mu_up)
	      state_next_box(2) = mu_up;  
	
	//!.....update firm-specific coordinates
	state_next_box_1(nfirm_count) = state_next_box(1);
	state_next_box_2(nfirm_count) = state_next_box(2);
	
	} //end FIRM LOOP

	state_next_box(3) = g_co_next;
	if(state_next_box(3) <= g_down)
	      state_next_box(3)= g_down;
	if(state_next_box(3) >= g_up)
	      state_next_box(3) = g_up;  
		
	double price_good_1;
	double price_good_2;
	
	double sum_good1, sum_good2;
	
	sum_good1 = 0.0;
	double gamma = 3.0;
	for(int i = 1; i <=nrfirms; i++){
	  sum_good1 = pow(e_current(i)*mu_temp_1(i),1.0 - gamma) + sum_good1;
	}
	price_good_1 = pow((1.0/nrfirms)*sum_good1,1.0/(1.0-gamma))*money;
	
	sum_good2 = 0.0;
	for(int i = 1; i <=nrfirms; i++){
	  sum_good2 = pow( e_current(i)*mu_temp_2(i),1.0 - gamma) + sum_good2;
	}
	price_good_2 = pow((1.0/nrfirms)*sum_good2,1.0/(1.0-gamma))*money;
	
	//!.....price level next period
	double p_co_true;
 	p_co_true = pow(0.5*pow(price_good_1,1.0 -theta) + 0.5*pow(price_good_2,1.0 - theta),1.0/(1.0 - theta))/money;	
	p_co_true = p_co_true/p_co_ss;
	
	money = money*g_co_next;
	
	p_co_next = p_co_true/g_co_next;
	
	if(with_m_shock == 0){
	  state_next_box(4) = 1;
	}
	else
	{
	  state_next_box(4) = p_co_next;
	}
	
	if(state_next_box(4) <= p_down){
	      state_next_box(4)= p_down;}
	      
	if(state_next_box(4) >= p_up){
	      state_next_box(4) = p_up;}

	pco_file << loop_count << " " << p_co_true << " " << g_co << endl;
	
//!=====================================================================	  
	  } //end simulation loop
//!=====================================================================	

      cout << "=============================================" << endl;
      cout << nrsim  << "  x " << nrfirms << " random points were drawn" << endl;
      cout << "                                             "<< endl;
      cout << price_1 << " x price 1 occured out of " << nrsim*nrfirms << endl;
      cout << price_2 << " x price 2 occured out of " << nrsim*nrfirms << endl;
      cout << price_3 << " x price 3 occured out of " << nrsim*nrfirms << endl;
      
      cout << "how often in state 1: " << state_sum_1 << endl;
      cout << "how often in state 2: " << state_sum_2 << endl;

      cout << "   on rank " << rank << endl;
      cout << "   SIMULATION DONE " << endl;
      cout << "=============================================" << endl;

//!=====================================================================	  

      //!.....close file	
      pco_file.close();
      
//!=====================================================================
      return 1;
}  
//!=====================================================================

  

  
  

