%**************************************************************************
% Supplemental Material to Brumm and Scheidegger (2017)
% "Using Adaptive Sparse Grids to Solve High-Dimensional Dynamic Models" 
%**************************************************************************
% PRICE STATISTICS
%**************************************************************************
% This script calculates price statistics for the menu costs models in
% Brumm and Scheidegger (2017). It builds on supplemental material to
% Midrigan (2011). It proceeds in five steps: 
% 1) Read price data.
% 2) Construct regular prices.
% 3) Plot price and regular price. (if enabled)
% 4) Calculate price statistics.
% 5) Print price statistics.
%**************************************************************************

clear all; clc; 

%% 1) Read price from file:
PriceChangeTable = readtable(['2goods_data_orig\PriceChange.txt'],'Delimiter',' ','ReadVariableNames',false);
%PriceChangeTable = readtable(['3goods_data_orig\PriceChange.txt'],'Delimiter',' ','ReadVariableNames',false);
pricechange=[PriceChangeTable{:,{'Var4'}} PriceChangeTable{:,{'Var7'}}];
pricechange=pricechange(1000:26000,:); 
prices(1,:)=[1 1];
for ii=2:size(pricechange,1)
   prices(ii,:)=prices(ii-1,:).*pricechange(ii,:); 
end
prices=round(prices,5);

%% 2) Construct regular prices:
oldprices = prices; 
[T,K] = size(prices);
share=sum(prices);
share = share/sum(share);
share = repmat(share,T,1);

wind = 5;        % # of months around current price used to compute modal price
cutoff = 1/3;    % of used less than cutoff times, count as temporary price
acc = .5;        % how many times should price be available for modal calculation to count?

temp=zeros(size(prices));  %0: unclassified, 1: permanent, 2: temporary
temp1=temp;               
                         
modalprice=zeros(T,K);
fract=zeros(T,K);

for i=1:K
    index=find(prices(:,i)>0);
     
    for j=index(1):index(end)
    if prices(j,i)>0
       junk=prices(max(j-wind,1):min(j+wind,T),i);
       junk=junk(junk>0);
  
    if length(junk)-1>acc*(2*wind+1)
       if all(junk==junk(1))    
          pmode=junk(1);
          fract(j,i)=1;   
       else
         [njunk,xjunk]=hist(junk,unique(junk));
         djunk=sortrows([vec(njunk),vec(xjunk)],1);
         pmode=djunk(end,2);
         fract(j,i)=djunk(end,1)/length(junk);   
      end
      modalprice(j,i)=pmode;
   end
   end
   end   
end

pR=zeros(T,K);
for i=1:K
    index=find(prices(:,i)>0);
for j=index(1):index(end)
      if j==1
          pR(j,i)=modalprice(j,i).*(modalprice(j,i)>0)+prices(j,i).*(modalprice(j,i)==0);
      else
          if modalprice(j,i)>0 
              pR(j,i)=modalprice(j,i).*(fract(j,i)>=cutoff&modalprice(j,i)==prices(j,i))+pR(j-1,i).*(~((fract(j,i)>=cutoff&modalprice(j,i)==prices(j,i))));
          else
              pR(j,i)=pR(j-1,i);
          end
      end
end
end

for i=1:K  % if p_t is mode, make p_t+1 /p_t-1 modes as well if they are the same as p_t
   for ss = 1:wind
       index=find(prices(2:end,i)==prices(1:end-1,i)&prices(2:end,i)~=0&prices(1:end-1,i)~=0)+1;  %index of dates where prices don't change
       nommatch=pR(index,i)~=pR(index-1,i)&pR(index-1,i)~=0&pR(index,i)~=0;
       pmatch=pR(index-1,i)==prices(index-1,i);
       cmatch=pR(index,i)==prices(index,i);
       pR(index(nommatch&cmatch)-1,i)=prices(index(nommatch&cmatch),i);
       pR(index(nommatch&pmatch),i)=prices(index(nommatch&pmatch)-1,i);
   end
end

%% 3) Plot price and regular price:
plotind=0;    
if plotind==1
prices(prices==0)=NaN;
pR(isnan(prices))=NaN;
pR(pR==0)=NaN;
i=1;
plot(prices(1:1000,i),'-b.','LineWidth',2,'MarkerSize',12);
hold on
plot(pR(1:1000,i),'r','LineWidth',2);
end

%% 4) Calculate price statistics:
prices=pR;
prices(oldprices==0)=0;
oldprices(prices==0)=0;
deltap = log(oldprices(2:end,:)./oldprices(1:end-1,:));
p99 = prctile(abs(deltap(oldprices(2:end,:)>0 & oldprices(1:end-1,:)>0 & abs(deltap)>0)),99);
deltap = deltap(oldprices(2:end,:)>0 & oldprices(1:end-1,:)>0 & abs(deltap)>0 & abs(deltap)<p99);
deltapR = log(prices(2:end,:)./prices(1:end-1,:));
pR99 = prctile(abs(deltapR(prices(2:end,:)>0 & prices(1:end-1,:)>0 & abs(deltapR)>0)),99);
deltapR = deltapR(prices(2:end,:)>0 & prices(1:end-1,:)>0 & abs(deltapR)>0 & abs(deltapR)<pR99 & prices(2:end,:)==oldprices(2:end,:)& prices(1:end-1,:)==oldprices(1:end-1,:));
size   = abs(log(oldprices(2:end,:)./oldprices(1:end-1,:)));
size   = size(oldprices(2:end,:)>0 & oldprices(1:end-1,:)>0 & size>0 & size<p99);
sizeR = abs(log(prices(2:end,:)./prices(1:end-1,:)));
sizeR = sizeR(prices(2:end,:)>0 & prices(1:end-1,:)>0 & sizeR>0  & sizeR<pR99 & prices(2:end,:)==oldprices(2:end,:)& prices(1:end-1,:)==oldprices(1:end-1,:));
change = abs(oldprices(2:end,:)-oldprices(1:end-1,:))>0; 
change = change(oldprices(2:end,:)>0&oldprices(1:end-1,:)>0);
changeR = abs(prices(2:end,:)-prices(1:end-1,:))>0; 
changeR = changeR(prices(2:end,:)>0&prices(1:end-1,:)>0);

%% 5) Print price statistics:
 fprintf('\n');fprintf('-----------------------------------------------------------');
 fprintf('\n');fprintf('-------------- STATISTICS OF PRICE CHANGES ----------------');fprintf('\n');
 fprintf('-----------------------------------------------------------');fprintf('\n');
 fprintf('Frequency of price changes                = %9.3f  \n', mean(change));
 fprintf('Frequency of regular price changes        = %9.3f  \n', mean(changeR));
 fprintf('-----------------------------------------------------------');fprintf('\n');
 fprintf('Mean size of price changes                = %9.3f  \n', mean(size));
 fprintf('Mean size of regular price changes        = %9.3f  \n', mean(sizeR));
 fprintf('-----------------------------------------------------------');fprintf('\n');
 fprintf('STD size of price changes                 = %9.3f  \n', std(size));
 fprintf('STD size of regular price changes         = %9.3f  \n', std(sizeR));
 fprintf('-----------------------------------------------------------');fprintf('\n');
 fprintf('Kurtosis of regular price changes         = %9.3f  \n', kurtosis(sizeR));
 fprintf('-----------------------------------------------------------');fprintf('\n');
 fprintf('-----------------------------------------------------------');fprintf('\n');