/* ************************************************************************
 * Supplemental Material to Brumm and Scheidegger (2017)
 * "Using Adaptive Sparse Grids to Solve High-Dimensional Dynamic Models" 
 * ************************************************************************
 * 
 * This script simulates the menu cost model of Section 5.
 * 
 * It proceeds in three steps:
 * 
 * 1) main.cpp calls the routine simulate(). The latter reads the 
 *    sparse grid value function data specified in the two files: 
 *    "2goods_data/MC_2goods_state_1.txt";
 *    "2goods_data/MC_2goods_state_2.txt";
 * 
 * 2) It simulates the menu cost model.
 * 
 * 3) It reports the price changes along the simulation in:
 *    "results/PriceChange.txt"; 
 * 
 ************************************************************************** */

#include <mpi.h>
#include <stdlib.h>
#include <random>
#include "incl/SparseGrid.h"
#include "incl/offline_interpol.h"   
#include "incl/parameters.h"

using namespace std;

//!=====================================================================
// mappings from sparse grid to economic domain
//!=====================================================================
//!.....[0,1]^d -> economic domain

extern "C" { 
  double box_to_cube_mapping_(double&, double&, int&, double&, double&, double&,
	                    double&, double&, double&);

}


//!=====================================================================

//!.....routine to simulate the MC model
int simulate();

//!=====================================================================

int main(int argc, char **args) {
  
	//!.....Initialize MPI
	MPI_Init (&argc, &args);
	
	//!.....Call main postprocessing routine to simulate MC model
	simulate();
	
	//!.....Finalize MPI environment.
	MPI_Finalize();

	return 0;
}

//----------------------------------------------------------------------


int simulate()
{
    
  
//!-------------------------------------------------------------

	//!.....filename of the data we want to simulate on
	char* filename1 = "2goods_data/MC_2goods_state_1.txt";
	char* filename2 = "2goods_data/MC_2goods_state_2.txt";

	//!.....Load data objects
	offline_interpol interpolate_fun_state_1, interpolate_fun_state_2; 
	interpolate_fun_state_1.LoadData(filename1); 
	cout << "opening data file  "<< filename1 <<  endl; 
	interpolate_fun_state_2.LoadData(filename2); 
	cout << "opening data file  "<< filename2 <<  endl; 
	
	//!.....files where results are written to	
	char* filename_PriceChange = "results/PriceChange.txt";
		
	//!.....open output
	ofstream PriceChange;
	PriceChange.open(filename_PriceChange);	
	PriceChange << "iteration  " <<  "  Price change 1  " <<  "  Price change 2 "<<  "  Reg Price change 1  " <<  "  Reg Price change 2 " <<  " Price Type "<< endl;

	//!....next period's state, vector of length dim, in [0,1]^d
	AdaptiveARRAY<double> state_next_box; state_next_box.redim(dim);
	AdaptiveARRAY<double> state_next_cube; state_next_cube.redim(dim);
	AdaptiveARRAY<double> mu_co; mu_co.redim(dim/2);
	AdaptiveARRAY<double> mu_temp; mu_temp.redim(dim/2);
	AdaptiveARRAY<double> mu_temp_old; mu_temp_old.redim(dim/2);
	AdaptiveARRAY<double> mu_reg; mu_reg.redim(dim/2);
	AdaptiveARRAY<double> mu_reg_old; mu_reg_old.redim(dim/2);
	AdaptiveARRAY<double> mu_co_next; mu_co_next.redim(dim/2);

	//!.....two random normally distributed number generators
	unsigned seed_m = 1;
	std::default_random_engine generator_m(seed_m); 
	//std::default_random_engine generator_m; //(no fixed seed)
	std::normal_distribution<double> distribution_m(0.0,sigma_m);

	unsigned seed_a = 2;
	std::default_random_engine generator_a(seed_a); 
	std::normal_distribution<double> distribution_a(0.0,sigma_a);	
	
//!.....uniform distributed [0..1]	
	unsigned seed_unif1 = 3;
	std::default_random_engine generator_unif(seed_unif1);
	std::uniform_real_distribution<double> distribution_unif(0.0,1.0);

	unsigned seed_unif2 = 4;
	std::default_random_engine generator_unif_2(seed_unif2);
	std::uniform_real_distribution<double> distribution_unif_2(0.0,1.0);	
	
	unsigned seed_unif3 = 5;
	std::default_random_engine generator_rand_unif(seed_unif3);
	std::uniform_real_distribution<double> distribution_rand_unif(0.0,1.0);		
	
//!.....Initialize price change parameters
    p1_old = 1.0; 
	
    eta_a1_old  = 1.0;
	eta_a2_old  = 1.0;
	
	e_plus_old  = 1.0;
	
	mu_temp_old(1) = 1.0;
	mu_temp_old(2) = 1.0; 
	
	mu_reg(1) = 1.0;
	mu_reg(2) = 1.0;
	
    mu_reg_old(1) = 1.0;
	mu_reg_old(2) = 1.0;

	//!.....state counter
	int state_sum_1 =0, state_sum_2=0;

	//!.....state 1 to start with. afterwards a random number.	
        state_in = 1;
	
        //!.....set initial point on [0..1]^d	
	state_next_box(1) = 1.5;
	state_next_box(2) = 1.5;
	state_next_box(3) = 0.5*(g_up + g_down);
	state_next_box(4) = 0.5*(p_up + p_down);
	
	//!.....transform to cube [0..1]^d
	box_to_cube_mapping_(state_next_box(1), state_next_cube(1),dim, mu_up, mu_down, g_up, g_down, p_up, p_down);
	
	//!=====================================================================	
	//!.....Major simulation loop
	//!=====================================================================		
	for(int loop_count = 1; loop_count <=nrsim; loop_count++)
	{
	  if (verbosity == 1)
		    cout << " simulation point no. " << loop_count <<  endl;
	   delta_mu1_old = state_next_box(1); 
	   
	  //!.....in which state are we now - interpolant
	  if(state_in == 1)
	  {
	    interpolate_fun_state_1.Interpolate(state_next_cube,fvalue);
	    VF1 = fvalue[0];
	    mu_R = fvalue[1];
	    VF2 = fvalue[2];
	    mu = fvalue[3];
	    VF3 = fvalue[4];
	    state_sum_1 = state_sum_1 + 1;
	  }

	  else if (state_in == 2)
	  {
	    interpolate_fun_state_2.Interpolate(state_next_cube,fvalue);
	    VF1 = fvalue[0];
	    mu_R = fvalue[1];
	    VF2 = fvalue[2];
	    mu = fvalue[3];
	    VF3 = fvalue[4];
    	    state_sum_2 = state_sum_2 + 1;
	  }
	  
//!.......value function choices	  
	  if (VF1 > max(VF2,VF3))
	  {
	     x1s = mu_R;
	     x2s = mu_R;
	     mu_temp(1) = mu_R;
	     mu_temp(2) = mu_R;
	     mu_reg(1) = mu_R;
	     mu_reg(2) = mu_R;
	     price_disc = 1.0;
	     price_cont = mu_R;
	     price_1 = price_1 + 1; 	    
	  }
	  else if (VF2 > max(VF1,VF3))
	  {
    	     x1s = state_next_box(1);
	     x2s = state_next_box(2);
	     mu_temp(1) = mu;
	     mu_temp(2) = mu;
	     mu_reg(1) = state_next_box(1);
	     mu_reg(2) = state_next_box(2);
	     price_disc = 2.0;
	     price_cont = mu;
	     price_2 = price_2 + 1;
	  }
	  else if (VF3 > max(VF1,VF2))
	  {
       	     x1s = state_next_box(1);
	     x2s = state_next_box(2);
	     mu_temp(1) = state_next_box(1);
	     mu_temp(2) = state_next_box(2);
	     mu_reg(1) = state_next_box(1);
	     mu_reg(2) = state_next_box(2);
	     price_disc = 3.0;
	     price_cont = 0.0;
     	     price_3 = price_3 + 1;
	  }
	  
	  //!.......update coords
	  mu_co(1) = x1s; 
	  mu_co(2) = x2s;
	  g_co = state_next_box(3);
	  p_co = state_next_box(4);
	  
	  //!......Update grid points for simulation	  
	  //!......draw random numbers
  	  eta_m = distribution_m(generator_m);
	  
	  double eta_a1_normal = distribution_a(generator_a);
	  double eta_a2_normal = distribution_a(generator_a);
	  double eta_a1_tilde;
	  double eta_a2_tilde;
	  
	  double eta_a1_uniform  = distribution_unif(generator_unif);
	  double eta_a2_uniform  = distribution_unif_2(generator_unif_2);
	  
	  //!.....first random var
	  if(eta_a1_uniform <= p_a)
	  {
	    eta_a1_tilde = eta_a1_normal;
	  }
	  else 
	  {
	   eta_a1_tilde = 0.0; 
	  }
	  
	   //!.....second random var
	  if(eta_a2_uniform <= p_a)
	  {
	    eta_a2_tilde = eta_a2_normal;
	  }
	  else 
	  {
	   eta_a2_tilde = 0.0; 
	  }
	  
	  double eta_mean = 0.5*(eta_a1_tilde + eta_a2_tilde); 

	  eta_a1 = eta_a1_tilde + xi*eta_mean;
	  eta_a2 = eta_a2_tilde + xi*eta_mean;
	  
	  //!......get e_current
	 if(state_in == 1)
	 {
	  e_current = e_bar; 
	 }
	 else if (state_in == 2)
	 {
	  e_current = 1.0;
	 }	  

	  //!...... draw random number in [0..1]
	 double rand_unif = distribution_unif(generator_rand_unif); 
	  
	 //!......get e_next and update state for next period
	 if(state_in == 1)
	 {
	  prob_v1 = rho;
	  prob_v2 = 1.0 -rho;
	  if(rand_unif < prob_v1)
	  {
	    state_plus1 = 1;
	    e_next = e_bar;
	  }
	  else if(rand_unif >= prob_v1)
	  {
	    state_plus1 = 2;
	    e_next = 1;
	  }
	 }
	 else if (state_in == 2)
	 {
          prob_v1 = alpha;
	  prob_v2 = 1.0 -alpha;
	  if(rand_unif < prob_v1)
	  {
	    state_plus1 = 1;
	    e_next = e_bar;
	  }
	  else if(rand_unif >= prob_v1)
	  {
	    state_plus1 = 2;
	    e_next = 1;
	  }
	 }
	  
	//!.....compute price changes p(t) / p(t-1)
	delta_a1_plus_old  = eta_a1_old;
	delta_a2_plus_old  = eta_a2_old;
	
	Gplus  = state_next_box(3);
	
	e_plus_current  = e_current;

	//!.....price change		
	delta_price  = (mu_temp(1)/mu_temp_old(1))*exp(delta_a1_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	
	delta_price2 = (mu_temp(2)/mu_temp_old(2))*exp(delta_a2_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	
	PriceChange << loop_count << "   " << delta_price << "   " << delta_price2;	

	//!.....regular price change		
	delta_price_R  = (mu_reg(1)/mu_reg_old(1))*exp(delta_a1_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	
	delta_price2_R = (mu_reg(2)/mu_reg_old(2))*exp(delta_a2_plus_old)*(e_plus_current/e_plus_old)*Gplus;
	
	PriceChange << "   " << delta_price_R << "   " << delta_price2_R << " " << price_disc << endl;	
	  
	//!.....update old values	
	e_plus_old  = e_plus_current;	
	
	eta_a1_old = eta_a1; 
        eta_a2_old = eta_a2;
	
        mu_temp_old(1) = mu_temp(1);
	mu_temp_old(2) = mu_temp(2);
	
	mu_reg_old(1) = mu_reg(1); 
	mu_reg_old(2) = mu_reg(2); 

	//!.....update state + grid point here...         	    
	state_in = state_plus1;

 	g_co_next = exp(rho_m*log(g_co) + eta_m);
 	mu_co_next(1) = (mu_co(1)*e_current)/(g_co_next*exp(eta_a1)*e_next);
 	mu_co_next(2) = (mu_co(2)*e_current)/(g_co_next*exp(eta_a2)*e_next);
 	p_co_next = exp(zeta_0 + zeta_1*log(g_co_next) + zeta_2*log(p_co));
	
	state_next_box(1) = mu_co_next(1);
	state_next_box(2) = mu_co_next(2);
	state_next_box(3) = g_co_next;
	state_next_box(4) = p_co_next;
	
	//!force simulation into the box
	if(state_next_box(1) <= mu_down)
	      state_next_box(1)= mu_down;
	if(state_next_box(1) >= mu_up)
	      state_next_box(1) = mu_up;
	  
	if(state_next_box(2) <= mu_down)
	      state_next_box(2)= mu_down;
	if(state_next_box(2) >= mu_up)
	      state_next_box(2) = mu_up;  
	
	if(state_next_box(3) <= g_down)
	      state_next_box(3)= g_down;
	if(state_next_box(3) >= g_up)
	      state_next_box(3) = g_up;  
	
	if(state_next_box(4) <= p_down)
	      state_next_box(4)= p_down;
	if(state_next_box(4) >= p_up)
	      state_next_box(4) = p_up;  
		
	//!.....transform to cube [0..1]^d
	box_to_cube_mapping_(state_next_box(1), state_next_cube(1),dim, mu_up, mu_down, g_up, g_down, p_up, p_down);

//!=====================================================================	  
	  } //end simulation loop
//!=====================================================================	  

      cout << "=============================================" << endl;
      cout << nrsim << "  random points were drawn" << endl;
      cout << "                                             "<< endl;
      cout << price_1 << " x price 1 occured out of " << nrsim << endl;
      cout << price_2 << " x price 2 occured out of " << nrsim << endl;
      cout << price_3 << " x price 3 occured out of " << nrsim << endl;
      
      cout << "how often in state 1: " << state_sum_1 << endl;
      cout << "how often in state 2: " << state_sum_2 << endl;
      
      cout << "   SIMULATION DONE " << endl;
      cout << "=============================================" << endl;

//!=====================================================================	  

      //!.....close file	
      PriceChange.close();
      
//!=====================================================================
      return 1;
}  
//!=====================================================================
  

  
  

